package com.rizzoweb.alarm;

import java.util.Calendar;
import java.util.Date;

/**
 * This class represents a schedule (repeating or one-time)
 * on which an alarm will trigger.
 */
public class AlarmSchedule implements Comparable, java.io.Serializable {

	private int minute = -1;
	private int hour = -1;
	private int dayOfMonth = -1;
	private int month = -1;
	private int dayOfWeek = -1;
	private int year = -1;
	private boolean isRelative;
	private boolean isRepetitive;
	private long alarmTime;
	private transient AlarmListener listener;
	private static final boolean debug = false;

	private void debug(String s) {
	  if (debug)
		System.out.println("[" + Thread.currentThread().getName() + "] AlarmSchedule: " + s);
	}

	/**
	 * Creates a new AlarmSchedule to fire on the given date.
	 *
	 * @param aDate The alarm date to be added.
	 * @param aListener The listener to be notified when the alarm fires.
	 * @exception PastDateException if the alarm date is in the past
	 *		(or less than 1 second after the current date).
	 */
	public AlarmSchedule(Date aDate, AlarmListener aListener) throws PastDateException {

		Calendar alarm = Calendar.getInstance();

		this.listener = aListener;
		alarm.setTime(aDate);
		minute = alarm.get(Calendar.MINUTE);
		hour = alarm.get(Calendar.HOUR_OF_DAY);
		dayOfMonth = alarm.get(Calendar.DAY_OF_MONTH);
		month = alarm.get(Calendar.MONTH);
		year = alarm.get(Calendar.YEAR);
		isRepetitive = false;
		isRelative = false;
		alarmTime = aDate.getTime();
		checkAlarmTime();
	}

	/**
	  * Creates a new AlarmSchedule.
	  *
	  * @param delay the alarm delay in minutes (relative to now).
	  * @param isRepetitive <code>true</code> if the alarm must be
	  * reactivated, <code>false</code> otherwise.
	  * @param listener the alarm listener.
	  * @exception PastDateException if the alarm date is in the past
	  * (or less than 1 second closed to the current date).
	  */
	public AlarmSchedule(int delay, boolean isRepetitive, AlarmListener listener)
	  throws PastDateException {
	  if (delay < 1) {
		throw new PastDateException();
	  }

	  minute = delay;
	  this.listener = listener;
	  this.isRepetitive = isRepetitive;
	  if (minute < 1)
		throw new IllegalArgumentException();

	  isRelative = true;
	  updateAlarmTime();
	}

	/**
	  * Creates a new AlarmSchedule.
	  *
	  * @param minute minute of the alarm. Allowed values 0-59.
	  * @param hour hour of the alarm. Allowed values 0-23.
	  * @param dayOfMonth day of month of the alarm (-1 if every
	  * day). This attribute is exclusive with
	  * <code>dayOfWeek</code>. Allowed values 1-31.
	  * @param month month of the alarm (-1 if every month). Allowed values
	  * 0-11 (0 = January, 1 = February, ...). <code>java.util.Calendar</code>
	  * constants can be used.
	  * @param dayOfWeek day of week of the alarm (-1 if every
	  * day). This attribute is exclusive with
	  * <code>dayOfMonth</code>. Allowed values 1-7 (1 = Sunday, 2 =
	  * Monday, ...). <code>java.util.Calendar</code> constants can  be used.
	  * @param year year of the alarm. When this field is not set
	  * (i.e. -1) the alarm is repetitive (i.e. it is rescheduled when
	  *  reached).
	  * @param listener the alarm listener.
	  * @return the AlarmSchedule.
	  * @exception PastDateException if the alarm date is in the past
	  * (or less than 1 second closed to the current date).
	  */
	public AlarmSchedule(int minute, int hour, int dayOfMonth, int month,
					  int dayOfWeek, int year, AlarmListener listener)
	  throws PastDateException {
	  this.minute = minute;
	  this.hour = hour;
	  this.dayOfMonth = dayOfMonth;
	  this.month = month;
	  this.dayOfWeek = dayOfWeek;
	  this.year = year;
	  this.listener = listener;
	  isRepetitive = (year == -1);
	  isRelative = false;
	  updateAlarmTime();
	  checkAlarmTime();
	}

	/**
	 * Updates the alarm time for repetitive alarms.
	 */
	protected void updateAlarmTime() {

		if (isRelative) {
			alarmTime = System.currentTimeMillis() + (minute * 60000);
			return;
		}

		Calendar now = Calendar.getInstance();
		Calendar alarm = (Calendar) now.clone();

		debug("now: " + now.getTime());

		if (year != -1) {
			alarm.set(Calendar.YEAR, year);
		}
		if (month != -1) {
			alarm.set(Calendar.MONTH, month);
		}
		if (hour != -1) {
			alarm.set(Calendar.HOUR_OF_DAY, hour);
		}
		if (minute != -1) {
			alarm.set(Calendar.MINUTE, minute);
		}
		alarm.set(Calendar.SECOND, 0);


		// Increments minute if now >= alarm (for every minute alarms)
		if (minute == -1) {
			alarm.add(Calendar.MINUTE, 1);
		}

		// Increments hour if now >= alarm (for every hour alarms)
		if (hour == -1 && minute != -1 && now.get(Calendar.MINUTE) >= minute) {
			alarm.add(Calendar.HOUR_OF_DAY, 1);
		}


		// Incrementes dayOfYear if now >= alarm (for everyday alarms)
		if (dayOfMonth == -1 && dayOfWeek == -1 && hour != -1 && minute != -1 && (now.get(Calendar.HOUR_OF_DAY) > hour || (now.get(Calendar.HOUR_OF_DAY) == hour && now.get(Calendar.MINUTE) >= minute))) {
			alarm.roll(Calendar.DAY_OF_YEAR, true);
		}

		// Incrementes year if now >= alarm (for monthly or yearly alarms)
		if (month != -1 && year == -1 && (now.get(Calendar.MONTH) > month || (now.get(Calendar.MONTH) == month && (now.get(Calendar.DAY_OF_MONTH) > dayOfMonth || (now.get(Calendar.DAY_OF_MONTH) == dayOfMonth && (now.get(Calendar.HOUR_OF_DAY) > hour || (now.get(Calendar.HOUR_OF_DAY) == hour && (now.get(Calendar.MINUTE) >= minute)))))))) {
			alarm.add(Calendar.YEAR, 1);
		}

		// Weekly alarms
		if (dayOfWeek != -1) {
			int deltaOfDay = (7 + (dayOfWeek - now.get(Calendar.DAY_OF_WEEK))) % 7;
			debug("deltaOfDay: " + deltaOfDay);
			if (deltaOfDay != 0) {
				alarm.add(Calendar.DAY_OF_YEAR, deltaOfDay);
			}
			// Incrementes week if now >= alarm
			else
				if (now.get(Calendar.HOUR_OF_DAY) > hour || (now.get(Calendar.HOUR_OF_DAY) == hour && now.get(Calendar.MINUTE) >= minute)) {
					alarm.add(Calendar.WEEK_OF_YEAR, 1);
				}
		}

		// Monthly alarms
		else
			if (dayOfMonth != -1) {
				alarm.set(Calendar.DAY_OF_MONTH, dayOfMonth);

				// Incrementes month if now >= alarm (for weekly alarms)
				if (month == -1 && (now.get(Calendar.DAY_OF_MONTH) > dayOfMonth || (now.get(Calendar.DAY_OF_MONTH) == dayOfMonth && (now.get(Calendar.HOUR_OF_DAY) > hour || (now.get(Calendar.HOUR_OF_DAY) == hour && now.get(Calendar.MINUTE) >= minute))))) {
					alarm.roll(Calendar.MONTH, true);
				}
			}
		debug("alarm: " + alarm.getTime());
		alarmTime = alarm.getTime().getTime();
	}

	/**
	  * Checks if my alarm time is valid.
	  *
	  * @exception PastDateException if the alarm date is in the past
	  *		(or less than 1 second after the current time).
	  */
	protected void checkAlarmTime() throws PastDateException {

		if (alarmTime - System.currentTimeMillis() <= 1000) {
			throw new PastDateException();
		}
	}

	public String toString() {

		if (year != -1) {
			return "Alarm at " + new Date(alarmTime);
		} else {
			StringBuffer sb = new StringBuffer("Alarm params");
			sb.append(" minute=");
			sb.append(minute);
			sb.append(" hour=");
			sb.append(hour);
			sb.append(" dayOfMonth=");
			sb.append(dayOfMonth);
			sb.append(" month=");
			sb.append(month);
			sb.append(" dayOfWeek=");
			sb.append(dayOfWeek);
			sb.append(" (next alarm date=" + new Date(alarmTime) + ")");
			return sb.toString();
		}
	}

	/**
	 * Compares the given AlarmSchedule with me for ordering.
	 *
	 * @param anObject The AlarmSchedule with which to compare.
	 * @return A negative integer, zero, or a positive integer as I am
	 *		less than, equal to, or greater than the given AlarmSchedule.
	 * @exception ClassCastException if the argument is not an AlarmSchedule
	 */
	public int compareTo(Object anObject) {

		long otherTime = ((AlarmSchedule) anObject).alarmTime;

		if (alarmTime < otherTime)
			return -1;
		if (alarmTime > otherTime)
			return 1;
		return 0;
	}

	/**
	  * Answers whether I am equivalent some other AlarmSchedule.
	  *
	  * @param obj The AlarmSchedule with which to compare.
	  * @return true if my alarmTime is the same as the argument's.
	  * @exception ClassCastException if anything other than an AlarmSchedule
	  *		is passed as the argument.
	  */
	public boolean equals(Object obj) {
		return this.alarmTime ==  ((AlarmSchedule)obj).alarmTime;
	}

	/**
	 * Simple accessor to my alarm time (System.currentTimeMillis() format)
	 */
	protected long getAlarmTime() {
		return alarmTime;
	}

	/**
	 * Simple accessor to my day of the month property.
	 */
	protected int getDayOfMonth() {
		return dayOfMonth;
	}

	/**
	 * Simple accessor to my day of the week property.
	 */
	protected int getDayOfWeek() {
		return dayOfWeek;
	}

	/**
	 * Simple accessor to my hour property.
	 */
	protected int getHour() {
		return hour;
	}

	/**
	 * Simple accessor to my minute property.
	 */
	protected int getMinute() {
		return minute;
	}

	/**
	 * Simple accessor to my month property.
	 */
	protected int getMonth() {
		return month;
	}

	/**
	 * Simple accessor to my year property.
	 */
	protected int getYear() {
		return year;
	}

	/**
	 * Answers whether or not I am a repeating schedule.
	 */
	public boolean isRepetitive() {
		return isRepetitive;
	}

	protected void notifyListener() {

		if (listener != null) {
			listener.handleAlarm(this);
		}
	}

	protected void setAlarmTime(long newAlarmTime) {
		alarmTime = newAlarmTime;
	}

	public void setListener(AlarmListener newListener) {
		this.listener = newListener;
	}
}
