package com.rizzoweb.alarm;
/**
 * This class manages the thread which sleeps until the next alarm.
 * Methods are synchronized to prevent interference fromthe AlarmWaiter thread
 * and external threads.
 */
public class AlarmWaiter implements Runnable {

	protected AlarmManager mgr;
	protected Thread thread;
	private long sleepUntil = -1;
	private boolean debug = false;
	private boolean shutdown = false;


	private void debug(String s) {
	  if (debug)
		System.out.println("[" + Thread.currentThread().getName() + "] AlarmWaiter: " + s);
	}



	/**
	  * Creates a new AlarmWaiter.
	  *
	  * @param isDaemon true if the waiter thread should run as a daemon.
	  * @param threadName the name of the waiter thread
	  */
	public AlarmWaiter(AlarmManager mgr, boolean isDaemon, String waiterName) {
	  this.mgr = mgr;

	  // start the thread
	  thread = new Thread(this, waiterName);
	  thread.setDaemon(isDaemon);
	  thread.start();
	}



	/**
	  * Updates the time to sleep.
	  *
	  * @param sleepUntil the new time to sleep until.
	  */
	public synchronized void update(long sleepUntil) {
	  this.sleepUntil = sleepUntil;
	  debug("Update for " + sleepUntil); // timeToSleep);
	  debug("I notify the thread to update its sleeping time");
	  notify();
	}



	/**
	  * Restarts the thread for a new time to sleep until.
	  *
	  * @param sleepUntil the new time to sleep until.
	  */
	public synchronized void restart(long sleepUntil) {
	  this.sleepUntil = sleepUntil;
	  notify();
	}



	/**
	  * Stops (destroy) the thread.
	  */
	public synchronized void stop() {
	  shutdown = true;
	  notify();
	}



	public synchronized void run() {

		debug("running");

		while (!shutdown) {
			try {
				// check if there's alarm here
				if (sleepUntil <= 0) {
					// no alarm here. So wait for a new alarm to come along.
					wait();
				} else {
					// Wait until the alarm is ready.
					long timeout = sleepUntil - System.currentTimeMillis();
					if (timeout > 0) {
						wait(timeout);
					}
				}

				// now that we've awakened again, check if an alarm is due (within
				// 1 second)
				if (sleepUntil >= 0 && (sleepUntil - System.currentTimeMillis() < 1000)) {
					// yes, an alarm is ready. Notify the listeners.
					sleepUntil = -1;
					debug("notifying listeners");
					mgr.notifyListeners();
				}

			} catch (InterruptedException e) {
				debug("interrupted");
			}
		}
		debug("stopping");
	}
}